@@ -173,7 +173,7 @@ if ($HIP_PLATFORM eq "amd") {
         $HIPLDFLAGS .= " --ld-path=" . get_normalized_path("$HIP_CLANG_PATH/lld-link.exe");
     }
 
-    # get Clang RT Builtin path 
+    # get Clang RT Builtin path
     $HIP_CLANG_RT_LIB = `$HIPCC --print-runtime-dir`;
     chomp($HIP_CLANG_RT_LIB);
 
@@ -218,10 +218,6 @@ if ($HIP_PLATFORM eq "amd") {
     exit (-1);
 }
 
-# Add paths to common HIP includes:
-$HIPCXXFLAGS .= " -isystem " . get_normalized_path("$HIP_INCLUDE_PATH");
-$HIPCFLAGS .= " -isystem " . get_normalized_path("$HIP_INCLUDE_PATH");
-
 my $compileOnly = 0;
 my $needCXXFLAGS = 0;  # need to add CXX flags to compile step
 my $needCFLAGS = 0;    # need to add C flags to compile step
@@ -237,7 +233,7 @@ my $printLDFlags = 0;       # print HIPLDFLAGS
 my $runCmd = 1;
 my $buildDeps = 0;
 my $hsacoVersion = 0;
-my $funcSupp = 0;      # enable function support
+my $funcSupp = 1;      # enable function support
 my $rdc = 0;           # whether -fgpu-rdc is on
 
 my @options = ();
