@@ -14,9 +14,6 @@
 #endif
 
 #if !defined(__HIPCC_RTC__)
-#if defined(__cplusplus)
-#include <algorithm>
-#endif
 #include <limits.h>
 #include <stdint.h>
 #ifdef __OPENMP_AMDGCN__
@@ -49,6 +46,17 @@
 #define __DEVICE_NOCE__ __DEVICE__
 #endif
 
+// Device library provides fast low precision and slow full-recision
+// implementations for some functions. Which one gets selected depends on
+// __CLANG_GPU_APPROX_TRANSCENDENTALS__ which gets defined by clang if
+// -ffast-math or -fgpu-approx-transcendentals are in effect.
+#pragma push_macro("__FAST_OR_SLOW")
+#if defined(__CLANG_GPU_APPROX_TRANSCENDENTALS__)
+#define __FAST_OR_SLOW(fast, slow) fast
+#else
+#define __FAST_OR_SLOW(fast, slow) slow
+#endif
+
 // A few functions return bool type starting only in C++11.
 #pragma push_macro("__RETURN_TYPE")
 #ifdef __OPENMP_AMDGCN__
@@ -59,7 +67,7 @@
 #else
 #define __RETURN_TYPE int
 #endif
-#endif
+#endif // __OPENMP_AMDGCN__
 
 #if defined (__cplusplus) && __cplusplus < 201103L
 // emulate static_assert on type sizes
@@ -161,420 +169,6 @@ uint64_t __make_mantissa(const char *__tagp __attribute__((nonnull))) {
 }
 
 // BEGIN FLOAT
-#if defined(__cplusplus)
-__DEVICE__
-int abs(int __x) {
-  int __sgn = __x >> (sizeof(int) * CHAR_BIT - 1);
-  return (__x ^ __sgn) - __sgn;
-}
-__DEVICE__
-long labs(long __x) {
-  long __sgn = __x >> (sizeof(long) * CHAR_BIT - 1);
-  return (__x ^ __sgn) - __sgn;
-}
-__DEVICE__
-long long llabs(long long __x) {
-  long long __sgn = __x >> (sizeof(long long) * CHAR_BIT - 1);
-  return (__x ^ __sgn) - __sgn;
-}
-#endif
-
-__DEVICE__
-float acosf(float __x) { return __ocml_acos_f32(__x); }
-
-__DEVICE__
-float acoshf(float __x) { return __ocml_acosh_f32(__x); }
-
-__DEVICE__
-float asinf(float __x) { return __ocml_asin_f32(__x); }
-
-__DEVICE__
-float asinhf(float __x) { return __ocml_asinh_f32(__x); }
-
-__DEVICE__
-float atan2f(float __x, float __y) { return __ocml_atan2_f32(__x, __y); }
-
-__DEVICE__
-float atanf(float __x) { return __ocml_atan_f32(__x); }
-
-__DEVICE__
-float atanhf(float __x) { return __ocml_atanh_f32(__x); }
-
-__DEVICE__
-float cbrtf(float __x) { return __ocml_cbrt_f32(__x); }
-
-__DEVICE__
-float ceilf(float __x) { return __builtin_ceilf(__x); }
-
-__DEVICE__
-float copysignf(float __x, float __y) { return __builtin_copysignf(__x, __y); }
-
-__DEVICE__
-float cosf(float __x) { return __ocml_cos_f32(__x); }
-
-__DEVICE__
-float coshf(float __x) { return __ocml_cosh_f32(__x); }
-
-__DEVICE__
-float cospif(float __x) { return __ocml_cospi_f32(__x); }
-
-__DEVICE__
-float cyl_bessel_i0f(float __x) { return __ocml_i0_f32(__x); }
-
-__DEVICE__
-float cyl_bessel_i1f(float __x) { return __ocml_i1_f32(__x); }
-
-__DEVICE__
-float erfcf(float __x) { return __ocml_erfc_f32(__x); }
-
-__DEVICE__
-float erfcinvf(float __x) { return __ocml_erfcinv_f32(__x); }
-
-__DEVICE__
-float erfcxf(float __x) { return __ocml_erfcx_f32(__x); }
-
-__DEVICE__
-float erff(float __x) { return __ocml_erf_f32(__x); }
-
-__DEVICE__
-float erfinvf(float __x) { return __ocml_erfinv_f32(__x); }
-
-__DEVICE__
-float exp10f(float __x) { return __ocml_exp10_f32(__x); }
-
-__DEVICE__
-float exp2f(float __x) { return __ocml_exp2_f32(__x); }
-
-__DEVICE__
-float expf(float __x) { return __ocml_exp_f32(__x); }
-
-__DEVICE__
-float expm1f(float __x) { return __ocml_expm1_f32(__x); }
-
-__DEVICE__
-float fabsf(float __x) { return __builtin_fabsf(__x); }
-
-__DEVICE__
-float fdimf(float __x, float __y) { return __ocml_fdim_f32(__x, __y); }
-
-__DEVICE__
-float fdividef(float __x, float __y) { return __x / __y; }
-
-__DEVICE__
-float floorf(float __x) { return __builtin_floorf(__x); }
-
-__DEVICE__
-float fmaf(float __x, float __y, float __z) {
-  return __builtin_fmaf(__x, __y, __z);
-}
-
-__DEVICE__
-float fmaxf(float __x, float __y) { return __builtin_fmaxf(__x, __y); }
-
-__DEVICE__
-float fminf(float __x, float __y) { return __builtin_fminf(__x, __y); }
-
-__DEVICE__
-float fmodf(float __x, float __y) { return __ocml_fmod_f32(__x, __y); }
-
-__DEVICE_NOCE__
-float frexpf(float __x, int *__nptr) {
-  return __builtin_frexpf(__x, __nptr);
-}
-
-__DEVICE__
-float hypotf(float __x, float __y) { return __ocml_hypot_f32(__x, __y); }
-
-__DEVICE__
-int ilogbf(float __x) { return __ocml_ilogb_f32(__x); }
-
-__DEVICE__
-__RETURN_TYPE __finitef(float __x) { return __builtin_isfinite(__x); }
-
-__DEVICE__
-__RETURN_TYPE __isinff(float __x) { return __builtin_isinf(__x); }
-
-__DEVICE__
-__RETURN_TYPE __isnanf(float __x) { return __builtin_isnan(__x); }
-
-__DEVICE__
-float j0f(float __x) { return __ocml_j0_f32(__x); }
-
-__DEVICE__
-float j1f(float __x) { return __ocml_j1_f32(__x); }
-
-__DEVICE_NOCE__
-float jnf(int __n, float __x) { // TODO: we could use Ahmes multiplication
-                                // and the Miller & Brown algorithm
-  //       for linear recurrences to get O(log n) steps, but it's unclear if
-  //       it'd be beneficial in this case.
-  if (__n == 0)
-    return j0f(__x);
-  if (__n == 1)
-    return j1f(__x);
-
-  float __x0 = j0f(__x);
-  float __x1 = j1f(__x);
-  for (int __i = 1; __i < __n; ++__i) {
-    float __x2 = (2 * __i) / __x * __x1 - __x0;
-    __x0 = __x1;
-    __x1 = __x2;
-  }
-
-  return __x1;
-}
-
-__DEVICE__
-float ldexpf(float __x, int __e) { return __builtin_amdgcn_ldexpf(__x, __e); }
-
-__DEVICE__
-float lgammaf(float __x) { return __ocml_lgamma_f32(__x); }
-
-__DEVICE__
-long long int llrintf(float __x) { return __ocml_rint_f32(__x); }
-
-__DEVICE__
-long long int llroundf(float __x) { return __builtin_roundf(__x); }
-
-__DEVICE__
-float log10f(float __x) { return __ocml_log10_f32(__x); }
-
-__DEVICE__
-float log1pf(float __x) { return __ocml_log1p_f32(__x); }
-
-__DEVICE__
-float log2f(float __x) { return __ocml_log2_f32(__x); }
-
-__DEVICE__
-float logbf(float __x) { return __ocml_logb_f32(__x); }
-
-__DEVICE__
-float logf(float __x) { return __ocml_log_f32(__x); }
-
-__DEVICE__
-long int lrintf(float __x) { return __ocml_rint_f32(__x); }
-
-__DEVICE__
-long int lroundf(float __x) { return __builtin_roundf(__x); }
-
-__DEVICE_NOCE__
-float modff(float __x, float *__iptr) {
-  float __tmp;
-#ifdef __OPENMP_AMDGCN__
-#pragma omp allocate(__tmp) allocator(omp_thread_mem_alloc)
-#endif
-  float __r =
-      __ocml_modf_f32(__x, (__attribute__((address_space(5))) float *)&__tmp);
-  *__iptr = __tmp;
-  return __r;
-}
-
-// FIXME need a c version of nanf
-#if defined(__cplusplus)
-__DEVICE__
-float nanf(const char *__tagp __attribute__((nonnull))) {
-  union {
-    float val;
-    struct ieee_float {
-      unsigned int mantissa : 22;
-      unsigned int quiet : 1;
-      unsigned int exponent : 8;
-      unsigned int sign : 1;
-    } bits;
-  } __tmp;
-  __static_assert_type_size_equal(sizeof(__tmp.val), sizeof(__tmp.bits));
-
-  __tmp.bits.sign = 0u;
-  __tmp.bits.exponent = ~0u;
-  __tmp.bits.quiet = 1u;
-  __tmp.bits.mantissa = __make_mantissa(__tagp);
-
-  return __tmp.val;
-}
-#endif
-
-__DEVICE__
-float nearbyintf(float __x) { return __ocml_nearbyint_f32(__x); }
-
-__DEVICE__
-float nextafterf(float __x, float __y) {
-  return __ocml_nextafter_f32(__x, __y);
-}
-
-__DEVICE__
-float norm3df(float __x, float __y, float __z) {
-  return __ocml_len3_f32(__x, __y, __z);
-}
-
-__DEVICE__
-float norm4df(float __x, float __y, float __z, float __w) {
-  return __ocml_len4_f32(__x, __y, __z, __w);
-}
-
-__DEVICE__
-float normcdff(float __x) { return __ocml_ncdf_f32(__x); }
-
-__DEVICE__
-float normcdfinvf(float __x) { return __ocml_ncdfinv_f32(__x); }
-
-__DEVICE_NOCE__
-float normf(int __dim,
-            const float *__a) { // TODO: placeholder until OCML adds support.
-  float __r = 0;
-  while (__dim--) {
-    __r += __a[0] * __a[0];
-    ++__a;
-  }
-
-  return __ocml_sqrt_f32(__r);
-}
-
-__DEVICE__
-float powf(float __x, float __y) { return __ocml_pow_f32(__x, __y); }
-
-__DEVICE__
-float powif(float __x, int __y) { return __ocml_pown_f32(__x, __y); }
-
-__DEVICE__
-float rcbrtf(float __x) { return __ocml_rcbrt_f32(__x); }
-
-__DEVICE__
-float remainderf(float __x, float __y) {
-  return __ocml_remainder_f32(__x, __y);
-}
-
-__DEVICE_NOCE__
-float remquof(float __x, float __y, int *__quo) {
-  int __tmp;
-#ifdef __OPENMP_AMDGCN__
-#pragma omp allocate(__tmp) allocator(omp_thread_mem_alloc)
-#endif
-  float __r = __ocml_remquo_f32(
-      __x, __y, (__attribute__((address_space(5))) int *)&__tmp);
-  *__quo = __tmp;
-
-  return __r;
-}
-
-__DEVICE__
-float rhypotf(float __x, float __y) { return __ocml_rhypot_f32(__x, __y); }
-
-__DEVICE__
-float rintf(float __x) { return __ocml_rint_f32(__x); }
-
-__DEVICE__
-float rnorm3df(float __x, float __y, float __z) {
-  return __ocml_rlen3_f32(__x, __y, __z);
-}
-
-__DEVICE__
-float rnorm4df(float __x, float __y, float __z, float __w) {
-  return __ocml_rlen4_f32(__x, __y, __z, __w);
-}
-
-__DEVICE_NOCE__
-float rnormf(int __dim,
-             const float *__a) { // TODO: placeholder until OCML adds support.
-  float __r = 0;
-  while (__dim--) {
-    __r += __a[0] * __a[0];
-    ++__a;
-  }
-
-  return __ocml_rsqrt_f32(__r);
-}
-
-__DEVICE__
-float roundf(float __x) { return __builtin_roundf(__x); }
-
-__DEVICE__
-float rsqrtf(float __x) { return __ocml_rsqrt_f32(__x); }
-
-__DEVICE__
-float scalblnf(float __x, long int __n) {
-  return (__n < INT_MAX) ? __builtin_amdgcn_ldexpf(__x, __n)
-                         : __ocml_scalb_f32(__x, __n);
-}
-
-__DEVICE__
-float scalbnf(float __x, int __n) { return __builtin_amdgcn_ldexpf(__x, __n); }
-
-__DEVICE__
-__RETURN_TYPE __signbitf(float __x) { return __builtin_signbitf(__x); }
-
-__DEVICE_NOCE__
-void sincosf(float __x, float *__sinptr, float *__cosptr) {
-  float __tmp;
-#ifdef __OPENMP_AMDGCN__
-#pragma omp allocate(__tmp) allocator(omp_thread_mem_alloc)
-#endif
-  *__sinptr =
-      __ocml_sincos_f32(__x, (__attribute__((address_space(5))) float *)&__tmp);
-  *__cosptr = __tmp;
-}
-
-__DEVICE_NOCE__
-void sincospif(float __x, float *__sinptr, float *__cosptr) {
-  float __tmp;
-#ifdef __OPENMP_AMDGCN__
-#pragma omp allocate(__tmp) allocator(omp_thread_mem_alloc)
-#endif
-  *__sinptr = __ocml_sincospi_f32(
-      __x, (__attribute__((address_space(5))) float *)&__tmp);
-  *__cosptr = __tmp;
-}
-
-__DEVICE__
-float sinf(float __x) { return __ocml_sin_f32(__x); }
-
-__DEVICE__
-float sinhf(float __x) { return __ocml_sinh_f32(__x); }
-
-__DEVICE__
-float sinpif(float __x) { return __ocml_sinpi_f32(__x); }
-
-__DEVICE__
-float sqrtf(float __x) { return __ocml_sqrt_f32(__x); }
-
-__DEVICE__
-float tanf(float __x) { return __ocml_tan_f32(__x); }
-
-__DEVICE__
-float tanhf(float __x) { return __ocml_tanh_f32(__x); }
-
-__DEVICE__
-float tgammaf(float __x) { return __ocml_tgamma_f32(__x); }
-
-__DEVICE__
-float truncf(float __x) { return __builtin_truncf(__x); }
-
-__DEVICE__
-float y0f(float __x) { return __ocml_y0_f32(__x); }
-
-__DEVICE__
-float y1f(float __x) { return __ocml_y1_f32(__x); }
-
-__DEVICE_NOCE__
-float ynf(int __n, float __x) { // TODO: we could use Ahmes multiplication
-                                // and the Miller & Brown algorithm
-  //       for linear recurrences to get O(log n) steps, but it's unclear if
-  //       it'd be beneficial in this case. Placeholder until OCML adds
-  //       support.
-  if (__n == 0)
-    return y0f(__x);
-  if (__n == 1)
-    return y1f(__x);
-
-  float __x0 = y0f(__x);
-  float __x1 = y1f(__x);
-  for (int __i = 1; __i < __n; ++__i) {
-    float __x2 = (2 * __i) / __x * __x1 - __x0;
-    __x0 = __x1;
-    __x1 = __x2;
-  }
-
-  return __x1;
-}
 
 // BEGIN INTRINSICS
 
@@ -582,10 +176,16 @@ __DEVICE__
 float __cosf(float __x) { return __ocml_native_cos_f32(__x); }
 
 __DEVICE__
-float __exp10f(float __x) { return __ocml_native_exp10_f32(__x); }
+float __exp10f(float __x) {
+  const float __log2_10 = 0x1.a934f0p+1f;
+  return __builtin_amdgcn_exp2f(__log2_10 * __x);
+}
 
 __DEVICE__
-float __expf(float __x) { return __ocml_native_exp_f32(__x); }
+float __expf(float __x) {
+  const float __log2_e = 0x1.715476p+0;
+  return __builtin_amdgcn_exp2f(__log2_e * __x);
+}
 
 #if defined OCML_BASIC_ROUNDED_OPERATIONS
 __DEVICE__
@@ -702,13 +302,13 @@ float __fsub_rn(float __x, float __y) { return __x - __y; }
 #endif
 
 __DEVICE__
-float __log10f(float __x) { return __ocml_native_log10_f32(__x); }
+float __log10f(float __x) { return __builtin_log10f(__x); }
 
 __DEVICE__
-float __log2f(float __x) { return __ocml_native_log2_f32(__x); }
+float __log2f(float __x) { return __builtin_amdgcn_logf(__x); }
 
 __DEVICE__
-float __logf(float __x) { return __ocml_native_log_f32(__x); }
+float __logf(float __x) { return __builtin_logf(__x); }
 
 __DEVICE__
 float __powf(float __x, float __y) { return __ocml_pow_f32(__x, __y); }
@@ -726,8 +326,428 @@ __DEVICE__
 float __sinf(float __x) { return __ocml_native_sin_f32(__x); }
 
 __DEVICE__
-float __tanf(float __x) { return __ocml_tan_f32(__x); }
+float __tanf(float __x) {
+  return __sinf(__x) * __builtin_amdgcn_rcpf(__cosf(__x));
+}
 // END INTRINSICS
+
+#if defined(__cplusplus)
+__DEVICE__
+int abs(int __x) {
+  return __builtin_abs(__x);
+}
+__DEVICE__
+long labs(long __x) {
+  return __builtin_labs(__x);
+}
+__DEVICE__
+long long llabs(long long __x) {
+  return __builtin_llabs(__x);
+}
+#endif
+
+__DEVICE__
+float acosf(float __x) { return __ocml_acos_f32(__x); }
+
+__DEVICE__
+float acoshf(float __x) { return __ocml_acosh_f32(__x); }
+
+__DEVICE__
+float asinf(float __x) { return __ocml_asin_f32(__x); }
+
+__DEVICE__
+float asinhf(float __x) { return __ocml_asinh_f32(__x); }
+
+__DEVICE__
+float atan2f(float __x, float __y) { return __ocml_atan2_f32(__x, __y); }
+
+__DEVICE__
+float atanf(float __x) { return __ocml_atan_f32(__x); }
+
+__DEVICE__
+float atanhf(float __x) { return __ocml_atanh_f32(__x); }
+
+__DEVICE__
+float cbrtf(float __x) { return __ocml_cbrt_f32(__x); }
+
+__DEVICE__
+float ceilf(float __x) { return __builtin_ceilf(__x); }
+
+__DEVICE__
+float copysignf(float __x, float __y) { return __builtin_copysignf(__x, __y); }
+
+__DEVICE__
+float cosf(float __x) { return __FAST_OR_SLOW(__cosf, __ocml_cos_f32)(__x); }
+
+__DEVICE__
+float coshf(float __x) { return __ocml_cosh_f32(__x); }
+
+__DEVICE__
+float cospif(float __x) { return __ocml_cospi_f32(__x); }
+
+__DEVICE__
+float cyl_bessel_i0f(float __x) { return __ocml_i0_f32(__x); }
+
+__DEVICE__
+float cyl_bessel_i1f(float __x) { return __ocml_i1_f32(__x); }
+
+__DEVICE__
+float erfcf(float __x) { return __ocml_erfc_f32(__x); }
+
+__DEVICE__
+float erfcinvf(float __x) { return __ocml_erfcinv_f32(__x); }
+
+__DEVICE__
+float erfcxf(float __x) { return __ocml_erfcx_f32(__x); }
+
+__DEVICE__
+float erff(float __x) { return __ocml_erf_f32(__x); }
+
+__DEVICE__
+float erfinvf(float __x) { return __ocml_erfinv_f32(__x); }
+
+__DEVICE__
+float exp10f(float __x) { return __ocml_exp10_f32(__x); }
+
+__DEVICE__
+float exp2f(float __x) { return __builtin_exp2f(__x); }
+
+__DEVICE__
+float expf(float __x) { return __builtin_expf(__x); }
+
+__DEVICE__
+float expm1f(float __x) { return __ocml_expm1_f32(__x); }
+
+__DEVICE__
+float fabsf(float __x) { return __builtin_fabsf(__x); }
+
+__DEVICE__
+float fdimf(float __x, float __y) { return __ocml_fdim_f32(__x, __y); }
+
+__DEVICE__
+float fdividef(float __x, float __y) { return __x / __y; }
+
+__DEVICE__
+float floorf(float __x) { return __builtin_floorf(__x); }
+
+__DEVICE__
+float fmaf(float __x, float __y, float __z) {
+  return __builtin_fmaf(__x, __y, __z);
+}
+
+__DEVICE__
+float fmaxf(float __x, float __y) { return __builtin_fmaxf(__x, __y); }
+
+__DEVICE__
+float fminf(float __x, float __y) { return __builtin_fminf(__x, __y); }
+
+__DEVICE__
+float fmodf(float __x, float __y) { return __ocml_fmod_f32(__x, __y); }
+
+__DEVICE_NOCE__
+float frexpf(float __x, int *__nptr) {
+  return __builtin_frexpf(__x, __nptr);
+}
+
+__DEVICE__
+float hypotf(float __x, float __y) { return __ocml_hypot_f32(__x, __y); }
+
+__DEVICE__
+int ilogbf(float __x) { return __ocml_ilogb_f32(__x); }
+
+__DEVICE__
+__RETURN_TYPE __finitef(float __x) { return __builtin_isfinite(__x); }
+
+__DEVICE__
+__RETURN_TYPE __isinff(float __x) { return __builtin_isinf(__x); }
+
+__DEVICE__
+__RETURN_TYPE __isnanf(float __x) { return __builtin_isnan(__x); }
+
+__DEVICE__
+float j0f(float __x) { return __ocml_j0_f32(__x); }
+
+__DEVICE__
+float j1f(float __x) { return __ocml_j1_f32(__x); }
+
+__DEVICE_NOCE__
+float jnf(int __n, float __x) { // TODO: we could use Ahmes multiplication
+                                // and the Miller & Brown algorithm
+  //       for linear recurrences to get O(log n) steps, but it's unclear if
+  //       it'd be beneficial in this case.
+  if (__n == 0)
+    return j0f(__x);
+  if (__n == 1)
+    return j1f(__x);
+
+  float __x0 = j0f(__x);
+  float __x1 = j1f(__x);
+  for (int __i = 1; __i < __n; ++__i) {
+    float __x2 = (2 * __i) / __x * __x1 - __x0;
+    __x0 = __x1;
+    __x1 = __x2;
+  }
+
+  return __x1;
+}
+
+__DEVICE__
+float ldexpf(float __x, int __e) { return __builtin_amdgcn_ldexpf(__x, __e); }
+
+__DEVICE__
+float lgammaf(float __x) { return __ocml_lgamma_f32(__x); }
+
+__DEVICE__
+long long int llrintf(float __x) { return __builtin_rintf(__x); }
+
+__DEVICE__
+long long int llroundf(float __x) { return __builtin_roundf(__x); }
+
+__DEVICE__
+float log10f(float __x) { return __builtin_log10f(__x); }
+
+__DEVICE__
+float log1pf(float __x) { return __ocml_log1p_f32(__x); }
+
+__DEVICE__
+float log2f(float __x) { return __FAST_OR_SLOW(__log2f, __ocml_log2_f32)(__x); }
+
+__DEVICE__
+float logbf(float __x) { return __ocml_logb_f32(__x); }
+
+__DEVICE__
+float logf(float __x) { return __FAST_OR_SLOW(__logf, __ocml_log_f32)(__x); }
+
+__DEVICE__
+long int lrintf(float __x) { return __builtin_rintf(__x); }
+
+__DEVICE__
+long int lroundf(float __x) { return __builtin_roundf(__x); }
+
+__DEVICE_NOCE__
+float modff(float __x, float *__iptr) {
+  float __tmp;
+#ifdef __OPENMP_AMDGCN__
+#pragma omp allocate(__tmp) allocator(omp_thread_mem_alloc)
+#endif
+  float __r =
+      __ocml_modf_f32(__x, (__attribute__((address_space(5))) float *)&__tmp);
+  *__iptr = __tmp;
+  return __r;
+}
+
+// FIXME need a c version of nanf
+#if defined(__cplusplus)
+__DEVICE__
+float nanf(const char *__tagp __attribute__((nonnull))) {
+  union {
+    float val;
+    struct ieee_float {
+      unsigned int mantissa : 22;
+      unsigned int quiet : 1;
+      unsigned int exponent : 8;
+      unsigned int sign : 1;
+    } bits;
+  } __tmp;
+  __static_assert_type_size_equal(sizeof(__tmp.val), sizeof(__tmp.bits));
+
+  __tmp.bits.sign = 0u;
+  __tmp.bits.exponent = ~0u;
+  __tmp.bits.quiet = 1u;
+  __tmp.bits.mantissa = __make_mantissa(__tagp);
+
+  return __tmp.val;
+}
+#endif
+
+__DEVICE__
+float nearbyintf(float __x) { return __builtin_nearbyintf(__x); }
+
+__DEVICE__
+float nextafterf(float __x, float __y) {
+  return __ocml_nextafter_f32(__x, __y);
+}
+
+__DEVICE__
+float norm3df(float __x, float __y, float __z) {
+  return __ocml_len3_f32(__x, __y, __z);
+}
+
+__DEVICE__
+float norm4df(float __x, float __y, float __z, float __w) {
+  return __ocml_len4_f32(__x, __y, __z, __w);
+}
+
+__DEVICE__
+float normcdff(float __x) { return __ocml_ncdf_f32(__x); }
+
+__DEVICE__
+float normcdfinvf(float __x) { return __ocml_ncdfinv_f32(__x); }
+
+__DEVICE_NOCE__
+float normf(int __dim,
+            const float *__a) { // TODO: placeholder until OCML adds support.
+  float __r = 0;
+  while (__dim--) {
+    __r += __a[0] * __a[0];
+    ++__a;
+  }
+
+  return __builtin_sqrtf(__r);
+}
+
+__DEVICE__
+float powf(float __x, float __y) { return __ocml_pow_f32(__x, __y); }
+
+__DEVICE__
+float powif(float __x, int __y) { return __ocml_pown_f32(__x, __y); }
+
+__DEVICE__
+float rcbrtf(float __x) { return __ocml_rcbrt_f32(__x); }
+
+__DEVICE__
+float remainderf(float __x, float __y) {
+  return __ocml_remainder_f32(__x, __y);
+}
+
+__DEVICE_NOCE__
+float remquof(float __x, float __y, int *__quo) {
+  int __tmp;
+#ifdef __OPENMP_AMDGCN__
+#pragma omp allocate(__tmp) allocator(omp_thread_mem_alloc)
+#endif
+  float __r = __ocml_remquo_f32(
+      __x, __y, (__attribute__((address_space(5))) int *)&__tmp);
+  *__quo = __tmp;
+
+  return __r;
+}
+
+__DEVICE__
+float rhypotf(float __x, float __y) { return __ocml_rhypot_f32(__x, __y); }
+
+__DEVICE__
+float rintf(float __x) { return __builtin_rintf(__x); }
+
+__DEVICE__
+float rnorm3df(float __x, float __y, float __z) {
+  return __ocml_rlen3_f32(__x, __y, __z);
+}
+
+__DEVICE__
+float rnorm4df(float __x, float __y, float __z, float __w) {
+  return __ocml_rlen4_f32(__x, __y, __z, __w);
+}
+
+__DEVICE_NOCE__
+float rnormf(int __dim,
+             const float *__a) { // TODO: placeholder until OCML adds support.
+  float __r = 0;
+  while (__dim--) {
+    __r += __a[0] * __a[0];
+    ++__a;
+  }
+
+  return __ocml_rsqrt_f32(__r);
+}
+
+__DEVICE__
+float roundf(float __x) { return __builtin_roundf(__x); }
+
+__DEVICE__
+float rsqrtf(float __x) { return __ocml_rsqrt_f32(__x); }
+
+__DEVICE__
+float scalblnf(float __x, long int __n) {
+  return (__n < INT_MAX) ? __builtin_amdgcn_ldexpf(__x, __n)
+                         : __ocml_scalb_f32(__x, __n);
+}
+
+__DEVICE__
+float scalbnf(float __x, int __n) { return __builtin_amdgcn_ldexpf(__x, __n); }
+
+__DEVICE__
+__RETURN_TYPE __signbitf(float __x) { return __builtin_signbitf(__x); }
+
+__DEVICE_NOCE__
+void sincosf(float __x, float *__sinptr, float *__cosptr) {
+  float __tmp;
+#ifdef __OPENMP_AMDGCN__
+#pragma omp allocate(__tmp) allocator(omp_thread_mem_alloc)
+#endif
+#ifdef __CLANG_CUDA_APPROX_TRANSCENDENTALS__
+  __sincosf(__x, __sinptr, __cosptr);
+#else
+  *__sinptr =
+      __ocml_sincos_f32(__x, (__attribute__((address_space(5))) float *)&__tmp);
+  *__cosptr = __tmp;
+#endif
+}
+
+__DEVICE_NOCE__
+void sincospif(float __x, float *__sinptr, float *__cosptr) {
+  float __tmp;
+#ifdef __OPENMP_AMDGCN__
+#pragma omp allocate(__tmp) allocator(omp_thread_mem_alloc)
+#endif
+  *__sinptr = __ocml_sincospi_f32(
+      __x, (__attribute__((address_space(5))) float *)&__tmp);
+  *__cosptr = __tmp;
+}
+
+__DEVICE__
+float sinf(float __x) { return __FAST_OR_SLOW(__sinf, __ocml_sin_f32)(__x); }
+
+__DEVICE__
+float sinhf(float __x) { return __ocml_sinh_f32(__x); }
+
+__DEVICE__
+float sinpif(float __x) { return __ocml_sinpi_f32(__x); }
+
+__DEVICE__
+float sqrtf(float __x) { return __builtin_sqrtf(__x); }
+
+__DEVICE__
+float tanf(float __x) { return __ocml_tan_f32(__x); }
+
+__DEVICE__
+float tanhf(float __x) { return __ocml_tanh_f32(__x); }
+
+__DEVICE__
+float tgammaf(float __x) { return __ocml_tgamma_f32(__x); }
+
+__DEVICE__
+float truncf(float __x) { return __builtin_truncf(__x); }
+
+__DEVICE__
+float y0f(float __x) { return __ocml_y0_f32(__x); }
+
+__DEVICE__
+float y1f(float __x) { return __ocml_y1_f32(__x); }
+
+__DEVICE_NOCE__
+float ynf(int __n, float __x) { // TODO: we could use Ahmes multiplication
+                                // and the Miller & Brown algorithm
+  //       for linear recurrences to get O(log n) steps, but it's unclear if
+  //       it'd be beneficial in this case. Placeholder until OCML adds
+  //       support.
+  if (__n == 0)
+    return y0f(__x);
+  if (__n == 1)
+    return y1f(__x);
+
+  float __x0 = y0f(__x);
+  float __x1 = y1f(__x);
+  for (int __i = 1; __i < __n; ++__i) {
+    float __x2 = (2 * __i) / __x * __x1 - __x0;
+    __x0 = __x1;
+    __x1 = __x2;
+  }
+
+  return __x1;
+}
+
+
 // END FLOAT
 
 // BEGIN DOUBLE
@@ -882,7 +902,7 @@ __DEVICE__
 double lgamma(double __x) { return __ocml_lgamma_f64(__x); }
 
 __DEVICE__
-long long int llrint(double __x) { return __ocml_rint_f64(__x); }
+long long int llrint(double __x) { return __builtin_rint(__x); }
 
 __DEVICE__
 long long int llround(double __x) { return __builtin_round(__x); }
@@ -903,7 +923,7 @@ __DEVICE__
 double logb(double __x) { return __ocml_logb_f64(__x); }
 
 __DEVICE__
-long int lrint(double __x) { return __ocml_rint_f64(__x); }
+long int lrint(double __x) { return __builtin_rint(__x); }
 
 __DEVICE__
 long int lround(double __x) { return __builtin_round(__x); }
@@ -953,7 +973,7 @@ double nan(const char *__tagp) {
 #endif
 
 __DEVICE__
-double nearbyint(double __x) { return __ocml_nearbyint_f64(__x); }
+double nearbyint(double __x) { return __builtin_nearbyint(__x); }
 
 __DEVICE__
 double nextafter(double __x, double __y) {
@@ -969,7 +989,7 @@ double norm(int __dim,
     ++__a;
   }
 
-  return __ocml_sqrt_f64(__r);
+  return __builtin_sqrt(__r);
 }
 
 __DEVICE__
@@ -1019,7 +1039,7 @@ __DEVICE__
 double rhypot(double __x, double __y) { return __ocml_rhypot_f64(__x, __y); }
 
 __DEVICE__
-double rint(double __x) { return __ocml_rint_f64(__x); }
+double rint(double __x) { return __builtin_rint(__x); }
 
 __DEVICE_NOCE__
 double rnorm(int __dim,
@@ -1092,7 +1112,7 @@ __DEVICE__
 double sinpi(double __x) { return __ocml_sinpi_f64(__x); }
 
 __DEVICE__
-double sqrt(double __x) { return __ocml_sqrt_f64(__x); }
+double sqrt(double __x) { return __builtin_sqrt(__x); }
 
 __DEVICE__
 double tan(double __x) { return __ocml_tan_f64(__x); }
@@ -1226,7 +1246,7 @@ __DEVICE__
 double __dsqrt_rz(double __x) { return __ocml_sqrt_rtz_f64(__x); }
 #else
 __DEVICE__
-double __dsqrt_rn(double __x) { return __ocml_sqrt_f64(__x); }
+double __dsqrt_rn(double __x) { return __builtin_sqrt(__x); }
 #endif
 
 #if defined OCML_BASIC_ROUNDED_OPERATIONS
@@ -1318,11 +1338,11 @@ double min(double __x, double __y) { return __builtin_fmin(__x, __y); }
 
 #if !defined(__HIPCC_RTC__) && !defined(__OPENMP_AMDGCN__)
 __host__ inline static int min(int __arg1, int __arg2) {
-  return std::min(__arg1, __arg2);
+  return __arg1 < __arg2 ? __arg1 : __arg2;
 }
 
 __host__ inline static int max(int __arg1, int __arg2) {
-  return std::max(__arg1, __arg2);
+  return __arg1 > __arg2 ? __arg1 : __arg2;
 }
 #endif // !defined(__HIPCC_RTC__) && !defined(__OPENMP_AMDGCN__)
 #endif
@@ -1330,5 +1350,6 @@ __host__ inline static int max(int __arg1, int __arg2) {
 #pragma pop_macro("__DEVICE_NOCE__")
 #pragma pop_macro("__DEVICE__")
 #pragma pop_macro("__RETURN_TYPE")
+#pragma pop_macro("__FAST_OR_SLOW")
 
 #endif // __CLANG_HIP_MATH_H__
